/* ============================================================================
   = DNS2Host                                                                 =
   = Version 1.0 (2008-02-26)                                                 =
   = (C) Copyright 2008 by Steffen Lange (steffen-lange@lange-email.de)       =
   = All Rights Reserved                                                      =
   ============================================================================ */

// How To Create a User-Defined Service:
// http://support.microsoft.com/kb/137890/EN-US/
// Windows Server 2003 Resource Kit Tools:
// http://www.microsoft.com/downloads/details.aspx?FamilyID=9D467A69-57FF-4AE7-96EE-B18C4790CFFD&displaylang=en

#pragma comment(lib, "ws2_32.lib")
#pragma comment(lib, "iphlpapi.lib")
#pragma comment(lib, "dnsapi.lib")

#include <cstdio>
#include <cstdlib>
#include <cstdarg>
#include <cstring>
#include <winsock2.h>
#include <windows.h>
#include <tchar.h>
#include <ws2tcpip.h>
#include <iphlpapi.h>
#include <windns.h>

void Print(const _TCHAR *szStr, ...) {
#ifdef _DEBUG
	_TCHAR pBuf[1024];
	va_list vaList;
	va_start(vaList, szStr);
	_vsntprintf(pBuf, sizeof(pBuf) / sizeof(_TCHAR) - 1, szStr, vaList);
	OutputDebugString(pBuf);
#endif
}

_TCHAR * GetComputername() {
	_TCHAR *szName = new _TCHAR[MAX_COMPUTERNAME_LENGTH + 1];
	DWORD dwSize;

	if (!GetComputerName(szName, &dwSize)) {
		delete[] szName;
		return NULL;
	}

	return szName;
}

_TCHAR * GetComputernameFromDNS(_TCHAR *szHost) {
	DWORD dwSize = MAX_COMPUTERNAME_LENGTH + 1;
	_TCHAR *szName = new _TCHAR[dwSize];

	if (!DnsHostnameToComputerName(szHost, szName, &dwSize)) {
		delete[] szName;
		return NULL;
	}

	return szName;
}

char * GetIPAddress(char *szHost) {
	char *szIP;
	struct hostent *pHostent;
	struct in_addr inAddr;

	if (!(pHostent = gethostbyname(szHost)))
		return NULL;

	inAddr.s_addr = *(u_long *)pHostent->h_addr_list[0];
	char *szTemp = inet_ntoa(inAddr);
	szIP = new char[strlen(szTemp) + 1];
	strcpy(szIP, szTemp);

	return szIP;
}

_TCHAR * GetReverseIP(_TCHAR *szIP) {
	_TCHAR *szZone = _T(".IN-ADDR.ARPA");
	_TCHAR *szRevIP = new _TCHAR[_tcslen(szIP) + _tcslen(szZone) + 1];
	int n1, n2, n3, n4;

	_stscanf(szIP, _T("%d.%d.%d.%d"), &n1, &n2, &n3, &n4);
	_stprintf(szRevIP, _T("%d.%d.%d.%d%s"), n4, n3, n2, n1, szZone);

	return szRevIP;
}

_TCHAR * GetDNSName(_TCHAR *szIP) {
	_TCHAR *szName;
	PIP4_ARRAY pServerList;
	FIXED_INFO *pFixedInfo;
	ULONG ulOutBufLen = sizeof(FIXED_INFO);
	PDNS_RECORD pDnsRecord;
	_TCHAR *pOwnerName;

	if (!(pServerList = (PIP4_ARRAY)LocalAlloc(LPTR, sizeof(IP4_ARRAY))))
		return NULL;

	pFixedInfo = (FIXED_INFO *)GlobalAlloc(GPTR, ulOutBufLen);
	if (GetNetworkParams(pFixedInfo, &ulOutBufLen) == ERROR_BUFFER_OVERFLOW) {
		GlobalFree(pFixedInfo);
		pFixedInfo = (FIXED_INFO *)GlobalAlloc(GPTR, ulOutBufLen);
	}

	if (!GetNetworkParams(pFixedInfo, &ulOutBufLen)) {
		IP_ADDR_STRING *pIPAddr = &pFixedInfo->DnsServerList;
		if (pIPAddr) {
			pServerList->AddrCount = 1;
			pServerList->AddrArray[0] = inet_addr(pIPAddr->IpAddress.String);
		}
	}

	GlobalFree(pFixedInfo);

	pOwnerName = GetReverseIP(szIP);
	int nRet = DnsQuery(pOwnerName, DNS_TYPE_PTR, DNS_QUERY_BYPASS_CACHE, pServerList, &pDnsRecord, NULL);
	delete[] pOwnerName;

	if (nRet) {
		LocalFree(pServerList);
		return NULL;
	}

	_TCHAR *szTemp = pDnsRecord->Data.PTR.pNameHost;
	szName = new _TCHAR[_tcslen(szTemp) + 1];
	_tcscpy(szName, szTemp);

	DnsRecordListFree(pDnsRecord, DnsFreeRecordListDeep);

	LocalFree(pServerList);

	return szName;
}

bool DoSystemReboot(_TCHAR *szMessage, DWORD dwTimeout) {
	HANDLE hToken;
	TOKEN_PRIVILEGES tkp;

	if (OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) {
		if (LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid)) {
			tkp.PrivilegeCount = 1;
			tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
			if (AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES)NULL, 0)) {
				dwTimeout = dwTimeout < MAX_SHUTDOWN_TIMEOUT ? dwTimeout : MAX_SHUTDOWN_TIMEOUT;
				return InitiateSystemShutdownEx(NULL, szMessage, dwTimeout, TRUE, TRUE, SHTDN_REASON_MAJOR_OTHER | SHTDN_REASON_MINOR_OTHER) ? true : false;
			}
		}
	}

	return false;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
	bool bRebootSystem = false;

	WSADATA wsaData;
	if (WSAStartup(MAKEWORD(2, 2), &wsaData))
		return EXIT_FAILURE;

	char *IP_ = GetIPAddress(NULL);
	_TCHAR *IP = new _TCHAR[strlen(IP_) + 1];
	mbstowcs(IP, IP_, strlen(IP_) + 1);
	Print(_T("IP Address: %s\n"), IP);

	_TCHAR *DNS = GetDNSName(IP);
	Print(_T("DNS Name: %s\n"), DNS);

	_TCHAR *Computer = GetComputername();
	Print(_T("NetBIOS Name: %s\n"), Computer);

	_TCHAR *Computer2 = GetComputernameFromDNS(DNS);
	Print(_T("New Name: %s\n"), Computer2);

	WSACleanup();

	if (Computer != NULL && Computer2 != NULL)
		if (_tcslen(Computer2) > 0 && _tcscmp(Computer, Computer2) != 0) {
			if (SetComputerNameEx(ComputerNamePhysicalDnsHostname, Computer2)) {
				bRebootSystem = true;
			}
		}

	if (IP_ != NULL) delete[] IP_;
	if (IP != NULL) delete[] IP;
	if (DNS != NULL) delete[] DNS;
	if (Computer != NULL) delete[] Computer;
	if (Computer2 != NULL) delete[] Computer2;

	if (bRebootSystem)
		DoSystemReboot(_T("The NetBIOS name and DNS host name of this machine have been changed."), 30);

	return EXIT_SUCCESS;
}
